#include "Peeling.h"
#include "Debug.h"
#include "mex.h"
#include <algorithm>
#include <cmath>
#include <iostream>
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wwrite-strings"

#include <igl/C_STR.h>

// Make assert act like assert without crashing matlab
#undef assert
#define assert( isOK ) ( (isOK) ? (void)0 : (void) mexErrMsgTxt(C_STR(__FILE__<<":"<<__LINE__<<": failed assertion `"<<#isOK<<"'"<<std::endl) ) )

// Edges of a tetrahedron
static const int EDGES_3[][2] = {
  {1,2},
  {2,0},
  {0,1},
  {3,0},
  {3,1},
  {3,2}
};

static void print(const std::vector<int> & v)
{
  using namespace std;
  for(int i = 0;i<v.size();i++)
  {
    mexPrintf("%d ",v[i]+1);
  }
  mexPrintf("\n");
}

// Given a list of element indices, find all those among them that have the
// same outside status as out_query. 
//
// Inputs:
//   outside  current indicator of outside elements
//   Ele list of element indices
//   out_query  query value
// Returns subset of Ele that are inside
static std::vector<int> find_match(
  const BoolVector & outside,
  const std::vector<int> & Ele,
  const bool out_query)
{
  using namespace std;
  vector<int> IEle;
  // Only keep those that are inside
  for(vector<int>::const_iterator nit = Ele.begin(); nit!=Ele.end(); nit++)
  {
    if(out_query == outside[*nit])
    {
      IEle.push_back(*nit);
    }
  }
  return IEle;
}

template<unsigned int d> void Peeling<d>::init(int numElements, const double *pElements, const double *pValues, int numFacets, const double *pFacets)
{
	m_elements.resize(numElements);
	m_values.resize(numElements);
	for (int i=0; i<numElements; i++)
	{
		int elem[d+1];
		for (int c=0; c<d+1; c++) elem[c] = (int)pElements[i + c*numElements] - 1; // MATLAB indexing fix
		m_elements[i].initArr(elem);

		m_values[i] = pValues[i];
		//m_elements[i] = Tuple<d+1>(pElements[i], pElements[i], pElements[i], pElements[i], );
	}

	m_constrFacets.resize(numFacets);
	for (int i=0; i<numFacets; i++)
	{
		int facet[d];
		for (int c=0; c<d; c++) facet[c] = (int)pFacets[i + c*numFacets] - 1; // MATLAB indexing fix
		m_constrFacets[i].initArr(facet);
	}

	bool ret = prepareFacetElementMap();
	if (!ret) mexErrMsgTxt("prepareFacetElementMap failed");	

	// test if all constrained facets are facets of the input elements
	ret = testConstrFacetsPresence();
	if (!ret) mexErrMsgTxt("some constrained facets are not element facets");	
	prepareConstrFacetSet();
	prepareElemNeighbors();		
  // Default graph cut parameters
  source_v = *std::min_element(m_values.begin(),m_values.end());
  dest_v = *std::max_element(m_values.begin(),m_values.end());
  //source_v = 0;
  //dest_v = pow(2.0,d-1)*M_PI;
  smooth_w = 1e4;
  data_w = 255;


  if(d==3)
  {
    prepareEdgeElementMap();
  }
  prepareVertexElementMap();
  
}

template<unsigned int d> void Peeling<d>::init(int numElements, const double *pElements, const double *pValues, int numFacets, const double *pFacets,
  const std::map<std::pair<int,int>,double> & pSmoothnessTerm,
  const std::vector<double> & pDataTermIn,
  const std::vector<double> & pDataTermOut)
{
  m_SmoothnessTerm = pSmoothnessTerm;
  m_DataTermIn = pDataTermIn;
  m_DataTermOut = pDataTermOut;
  return init(numElements, pElements, pValues, numFacets, pFacets);
}

template<unsigned int d> void Peeling<d>::peel(BoolVector &outside)
{
	DoubleIntPairVector sortedElements = sortElementsByValues();
	outside.resize(getNumElements());
	for (int e=0; e<getNumElements(); e++) outside[e] = false;
	//outside.assign(false, getNumElements()); // all inside

	int cnt;
	for (int step=0; step<400; step++)
	{
		cnt = 0;		
		for (int k=0; k<getNumElements(); k++)
		{
			const int e = sortedElements[k].second;
			if (outside[e]) continue;

			// can e-th element be put outside?
			bool putOutside = true; // let's be optimistic and assume we can put element e outside

			// if there is a constrained facet and its other element is outside, we cannot put e outside:
			for (int i=0; i<d+1; i++)
			{
				if (m_elemConstrNeighbors[e].getV(i))
				{
					const int neighborIdx = m_elemNeighbors[e].getV(i);
					if (neighborIdx == -1) putOutside = false; // if constrained facet has no neighbor, then we can't put e outside
					else
					{
						assert(neighborIdx >= 0 && neighborIdx < getNumElements());
						if (outside[neighborIdx]) putOutside = false;
					}
				}
			}

			if (!putOutside) continue;
			// second test: if e is marked as outside, will the inside elements still form a connected component?		

			outside[e] = true; // temporarily mark as outside and see what happens

			// find some vertex inside:
			int insideVertex = -1, numInsideVerts = 0;
			for (int i=0; i<getNumElements(); i++)
			{
				if (!outside[i]) 
				{
					insideVertex = i;
					numInsideVerts++;
				}
			}
			assert(insideVertex >= 0);
			assert(insideVertex != e);

			int sizeCC = connectedComponentInside(outside, insideVertex);
			if (sizeCC < numInsideVerts) 
			{
				putOutside = false; // connected component smaller -> something has been cut off
			}

			outside[e] = putOutside;

			if (putOutside) cnt++;				
		}
		TRACE("step %i: %i elements peeled out\n", step, cnt);
		if (cnt == 0) break;
	}	
}

template<unsigned int d> void Peeling<d>::peelOneByOne(BoolVector &outside, bool peelFromOutside,
  const bool singleCC,
  const bool minCutCapacity)
{
	DoubleIntPairVector sortedElements = sortElementsByValues();
	outside.assign(getNumElements(), false); // all inside
	
	for (int s=0; s<getNumElements(); s++)
	{
		int elementToPeel = -1;		
		for (int k=0; k<getNumElements(); k++)
		{
			const int e = sortedElements[k].second;
			if (outside[e]) continue;

			// can e-th element be put outside?
			bool putOutside = true; // let's be optimistic and assume we can put element e outside

			if (peelFromOutside)
			{	// perform check whether element e is either adjacent to the boundary or to another element outside
				bool passedPFOtest = false;
				for (int i=0; i<d+1; i++)
				{
					const int neighborIdx = m_elemNeighbors[e].getV(i);
					if (neighborIdx == -1) passedPFOtest = true; // is adjacent to boundary
					else
					{
						assert(neighborIdx >= 0 && neighborIdx < getNumElements());
						if (outside[neighborIdx]) passedPFOtest = true; // is adjacent to element outside
					}
				}	
				if (!passedPFOtest)
				{
					putOutside = false;
					continue;
				}
			}

			// constrained facets test: if there is a constrained facet and its other element is outside, we cannot put e outside:
			for (int i=0; i<d+1; i++)
			{
				if (m_elemConstrNeighbors[e].getV(i))
				{
					const int neighborIdx = m_elemNeighbors[e].getV(i);
					if (neighborIdx == -1) putOutside = false; // if constrained facet has no neighbor, then we can't put e outside
					else
					{
						assert(neighborIdx >= 0 && neighborIdx < getNumElements());
						if (outside[neighborIdx]) putOutside = false;
					}
				}
			}

			if (!putOutside) continue;
			// final test: if e is marked as outside, will the inside elements still form a connected component?		

			assert(outside[e] == false);
			outside[e] = true; // temporarily mark as outside and see what happens

			// find some vertex inside:
			int insideVertex = -1, numInsideVerts = 0;
			for (int i=0; i<getNumElements(); i++)
			{
				if (!outside[i]) 
				{
					insideVertex = i;
					numInsideVerts++;
				}
			}
			assert(insideVertex >= 0);
			assert(insideVertex != e);

			int sizeCC = connectedComponentInside(outside, insideVertex);
			if (sizeCC < numInsideVerts) 
			{
				putOutside = false; // connected component smaller -> something has been cut off
			}
			outside[e] = false; // revert

			if (putOutside) 
			{
				elementToPeel = e;
				break;
			}
                }
		

		if (elementToPeel == -1)
		{
			TRACE("Finished after peeling %i elements\n", s);
			break;
		}

		TRACE("step %i, peeling element %i\n", s, elementToPeel);
		assert(elementToPeel >= 0);
		outside[elementToPeel] = true;
	}	
}

template<unsigned int d> int Peeling<d>::improve(
  BoolVector &outside,
  std::queue<int> & dirtyQ,
  const bool must_be_legal)
{
  using namespace std;
  int flips = 0;
  while(dirtyQ.size()>0)
  {
    const int e = dirtyQ.front();
    dirtyQ.pop();

    double change, cur_cost, new_cost;
    change = cut_capacity_change(outside,e,!outside[e]);
    //cur_cost = cut_capacity(outside);
    //// toggle e
    //outside[e] = !outside[e];
    //new_cost = cut_capacity(outside);
    //// toggle e
    //outside[e] = !outside[e];
    //if( fabs( (new_cost-cur_cost)-change) > 1)
    //{
    //  cout<<"new - cur = "<<new_cost<<" - "<<cur_cost<<" = "<<new_cost-cur_cost<<
    //    " ?= "<<change<<endl;
    //  mexErrMsgTxt("change differs");
    //}

    // Consider both neighbors of facet, which increases the energy the least?
    // Add to inside accordingly
    //if( cost(outside,e,!outside[e]) < cost(outside,e,outside[e]))
    if(change<0)
    {

      // Now check that it would be legal to flip
      bool legal = is_legal(outside,e,!outside[e]);
      if(!legal && must_be_legal)
      {
        //cout<<"SKIPPED ILLEGAL IMPROVEMENT"<<endl;
        continue;
      }

      outside[e] = !outside[e];
#ifdef SEQ
      cout<<"  "<<e+1<<" "<<outside[e]<<" % "<<__FUNCTION__<<endl;
#endif
      flips++;
      // now all e's neighbors are also dirty
      for(int ni = 0;ni<(d+1);ni++)
      {
        const int n = m_elemNeighbors[e].getV(ni);
        // Only consider real neighbors
        if(n != -1)
        {
          dirtyQ.push(n);
          // Q: Is this ever happening?
          // A: Yes.
        }
      }
    }
  }
  return flips;
}

template<unsigned int d> 
void Peeling<d>::project(BoolVector &outside)
{
  project_constrained_facets(outside);
#ifdef MANIFOLD_CONSTRAINT
  if(d==3)
  {
    project_edge_manifold(outside);
  }
  project_vertex_manifold(outside);
#endif
}

template<unsigned int d> 
void Peeling<d>::project_constrained_facets(BoolVector &outside)
{
  using namespace std;
  // Find all unconvered input faces
  vector<Tuple<d> > UF;
  // loop over facets
  const int numConstrFacets = (int)m_constrFacets.size();
  for (int i=0; i<numConstrFacets; i++)
  {
    Tuple<d> facet = m_constrFacets[i];
    facet.sort();
    assert(m_facetElementMap.count(facet)==1);
    Tuple<2> record = m_facetElementMap[facet];
    assert(record.getV(0)!=-1);
    if(
     outside[record.getV(0)] &&
     (record.getV(0) == -1 || outside[record.getV(0)]))
    {
      UF.push_back(facet);
    }
  }
  // UF now contains all uncovered input faces (each sorted)

  // Loop over each uncovered face
  // Q: Currently the order is arbitrary. Is there an optimal ordering? Would a
  // greedy ordering improve things?
  // A: How about instead. Pick the unsatisfied facet whose hard constraint
  // (however its chosen) will make minimal change to the graph cut energy? 
  // This implies a dynamic priority queue or a naive O(#UF^2)
  for(typename vector<Tuple<d> >::const_iterator ufit = UF.begin();
    ufit != UF.end();
    ufit++)
  {
    const Tuple<d> & facet = *ufit;
    // Get neighboring element indices
    assert(m_facetElementMap.count(facet)==1);
    Tuple<2> record = m_facetElementMap[facet];
    // First check that we really need to process this, since adjacent elements
    // may have been added by previous iteration
    const int a = record.getV(0);
    const int b = record.getV(1);
    if(!outside[a] || (b!=-1 && !outside[b]))
    {
      // safe to skip
      continue;
    }

    //// Consider both neighbors of facet, which contributes smallest change
    //const int e = 
    //  //b==-1?a:(cost(outside,a,false)<cost(outside,b,false)?a:b);
    //  b==-1?a:
    //    (cut_capacity_change(outside,a,false)<
    //       cut_capacity_change(outside,b,false)?a:b);
    //// Add to inside accordingly
    //outside[e] = false;

    // Consider both neighbors of facet, which has largest winding number
    // (value)
    const int e = b==-1?a:(m_values[a] > m_values[b]?a:b);
    // Add to inside accordingly
    assert(outside[e]);
    outside[e] = false;
#ifdef SEQ
    cout<<"  "<<e+1<<" "<<outside[e]<<" % "<<__FUNCTION__<<endl;
#endif

    // Now all neighbors of e are "dirty" meaning changing their status in
    // outside could possibly improve the overall cost
    queue<int> dirtyQ;
    for(int ni = 0;ni<(d+1);ni++)
    {
      const int n = m_elemNeighbors[e].getV(ni);
      // Only consider real neighbors
      if(n != -1)
      {
        dirtyQ.push(n);
      }
    }
    // Try to improve things
    const int flips = improve(outside,dirtyQ,true);
    //mexPrintf("skipping improve...\n");
    //cout<<"flips: "<<flips<<endl;

  }

}

template<unsigned int d> 
void Peeling<d>::project_edge_manifold(BoolVector &outside)
{
  using namespace std;
  // loop over edges
  for(
    EdgeElementMap::const_iterator eit = m_edgeElementMap.begin();
    eit != m_edgeElementMap.end();
    eit++)
  {
    project_single_cc(eit->second,outside);
  }
}

template<unsigned int d> 
void Peeling<d>::project_vertex_manifold(BoolVector &outside)
{
  using namespace std;
  // loop over vertices
  for(
    VertexElementMap::const_iterator vit = m_vertexElementMap.begin();
    vit != m_vertexElementMap.end();
    vit++)
  {
    //cout<<"%%  v: "<<vit->first+1<<endl;
    project_single_cc(vit->second,outside);
  }
}

template<unsigned int d> 
void Peeling<d>::project_single_cc(
  const std::vector<int> & Ele,
  BoolVector &outside)
{
  //project_single_cc_bfs(Ele,outside);
  project_single_cc_blind(Ele,outside);
}

template<unsigned int d> 
void Peeling<d>::project_single_cc_blind(
  const std::vector<int> & Ele,
  BoolVector &outside)
{
  using namespace std;

  // Put all outside elements into priority queue
  priority_queue<pair<double,int> > Q;
  for(vector<int>::const_iterator eit = Ele.begin(); eit!=Ele.end(); eit++)
  {
    if(outside[*eit])
    {
      Q.push(pair<double,int>(m_values[*eit],*eit));
      //Q.push(pair<double,int>(-cut_capacity_change(outside,*eit,false),*eit));
    }
  }
  const bool resort = false;
  //const bool resort = true;

  // Until both in and out are single CCs
  while(!(single_cc(find_match(outside,Ele,false)) && 
    single_cc(find_match(outside,Ele,true))))
  {

    if(resort)
    {
      queue<int> L;
      while(!Q.empty())
      {
        L.push(Q.top().second);
        Q.pop();
      }
      while(!L.empty())
      {
        Q.push(pair<double,int>(-cut_capacity_change(outside,L.front(),false),L.front()));
        L.pop();
      }
    }

    int e; 
    do{
      assert(!Q.empty());
      const pair<double,int> pe = Q.top();
      e = pe.second;
      Q.pop();
    }while(!outside[e]);

    assert(outside[e]);
    outside[e] = false;
#ifdef SEQ
    cout<<"  "<<e+1<<" "<<outside[e]<<" % "<<__FUNCTION__<<endl;
#endif
    // Now all neighbors of e are "dirty" meaning changing their status in
    // outside could possibly improve the overall cost
    queue<int> dirtyQ;
    for(int ni = 0;ni<(d+1);ni++)
    {
      const int n = m_elemNeighbors[e].getV(ni);
      // Only consider real neighbors
      if(n != -1)
      {
        dirtyQ.push(n);
      }
    }
    // Try to improve things
    const int flips = improve(outside,dirtyQ,true);
    //cout<<"flips: "<<flips<<endl;
  }
}

template<unsigned int d> 
void Peeling<d>::project_single_cc_bfs(
  const std::vector<int> & Ele,
  BoolVector &outside)
{
  using namespace std;
  // Annoying base case
  if(Ele.size() == 0)
  {
    return;
  }
  // assure yourself that the input really is one connected component
  assert(single_cc(Ele));

  // map of elements in Ele
  map<int,bool> in_Ele;
  map<int,bool> orig_out;
  map<int,bool> seen;
  int num_orig_in = 0;
  for(vector<int>::const_iterator eit = Ele.begin(); eit!=Ele.end(); eit++)
  {
    in_Ele[*eit] = true;
    seen[*eit] = false;
    orig_out[*eit] = outside[*eit];
    if(!outside[*eit])
    {
    }
    num_orig_in += (outside[*eit]?0:1);
  }

  queue<int> Q;
  // Arbitrary inside root
  for(vector<int>::const_iterator eit = Ele.begin(); eit!=Ele.end(); eit++)
  {
    if(!outside[*eit])
    {
      Q.push(*eit);
      break;
    }
  }

  // Annoying base case where nothing is inside
  // Strictly this is not a single connected component 
  if(Q.empty())
  {
    assert(num_orig_in== 0);
    return;
  }

  assert(Q.size() == 1);
  // Keep track of the original inside components we've seen
  int orig_in_count = 0;
  priority_queue< pair<double,int> > outQ;
  bool bad = false;

  // BFS until we see all original elements. Preferring to travel along
  // original elements
  while(!Q.empty())
  {
    const int e = Q.front();
    Q.pop();
    if(!seen[e])
    {
      seen[e] = true;
      // Keep track of whether we're seeing one of the original inside elements
      if(!orig_out[e])
      {
        orig_in_count++;
        if(orig_in_count == num_orig_in)
        {
          // We've seen all original inside elements
          return;
        }
      }
      // loop over neighbors of e
      for(int ni = 0;ni<(d+1);ni++)
      {
        const int n = m_elemNeighbors[e].getV(ni);
        // Only consider real neighbors in Ele
        if(n != -1 && in_Ele.count(n)==1)
        {
          assert(in_Ele[n]);
          // Only keep unseen neighbors
          if(!seen[n])
          {
            if(outside[n])
            {
              outQ.push(pair<double,int>(m_values[n],n));
              //outQ.push(pair<double,int>(-cut_capacity_change(outside,n,false),n));
            }else
            {
              Q.push(n);
            }
          }
        }
      }
    }

    // If there's no more inside elements to see then pop an outside element,
    // flip it to be inside and continue
    if(Q.empty())
    {
      assert(!outQ.empty());
      //bad = true;
      //{
      //  priority_queue< pair<double,int> > temp;
      //  cout<<"%% outQ: ";
      //  while(!outQ.empty())
      //  {
      //    const pair<double,int> p = outQ.top();
      //    outQ.pop();
      //    temp.push(p);
      //    cout<<"["<<p.first<<","<<p.second+1<<"] ";
      //  }
      //  cout<<endl;
      //  outQ = temp;
      //}

      // Again, outQ could/should be priority queue based on energy
      pair<double,int> po;
      // Note: Can be that some elements in outQ are now inside because of calls to
      // improve

      po = outQ.top();
      outQ.pop();
      const int o = po.second;
      if(outside[o])
      {
        // Found a truly outside frontier neighbor 
        outside[o] = false;
#ifdef SEQ
          cout<<"  "<<o+1<<" "<<outside[o]<<" % "<<__FUNCTION__<<endl;
#endif
        Q.push(o);
        // Now all neighbors of o are "dirty" meaning changing their status in
        // outside could possibly improve the overall cost
        queue<int> dirtyQ;
        for(int ni = 0;ni<(d+1);ni++)
        {
          const int n = m_elemNeighbors[e].getV(ni);
          // Only consider real neighbors
          if(n != -1)
          {
            dirtyQ.push(n);
          }
        }
        // Try to improve things
        const int flips = improve(outside,dirtyQ,true);
      }else
      {
        // Found a frontier neighbor who's been marked as in (by improve)
        // We might have seen this already by now. But that's OK
        Q.push(o);
        // Now we have something in Q so we can continue
      }

    }
    assert(!Q.empty());
  }

}

//template <unsigned int d>
//double Peeling<d>::cost(
//  const BoolVector & outside, 
//  const int e, 
//  const bool out)
//{
//  // Cut-capacity "cost" is only counted going from a source "inside" node to a
//  // dest "outside" node.
//
//  assert(e>=0);
//  assert(e<outside.size());
//  // Compute smoothness contribution
//  double smoothness = 0;
//  // loop over neighbors of e
//  for(int ni = 0;ni<(d+1);ni++)
//  {
//    const int n = m_elemNeighbors[e].getV(ni);
//    // Only consider real neighbors
//    if(n != -1)
//    {
//      // only punish when toggle would cut graph-cut graph edge
//      if( out != outside[n] )
//      {
//        // cuts only one way
//        smoothness += 2.0*(round(smooth_w*exp(-fabs(m_values[e]-m_values[n])))+1);
//      }
//    }
//  }
//  // Compute data contribution
//  double data = 0;
//  if(out)
//  {
//    // if out then we're cutting the edge from source to e
//    data += round(data_w*pow((m_values[e]-source_v)>0?(m_values[e]-source_v):0,2));
//  }else
//  {
//    // if in then we're cutting the edge from e to dest
//    data += round(data_w*pow((dest_v-m_values[e])>0?(dest_v-m_values[e]):0,2));
//  }
//  // total is weighted sum
//  return data + smoothness;
//}

template<unsigned int d> 
double Peeling<d>::cut_capacity(const BoolVector & outside)
{
  double data = 0;
  double smoothness = 0;
  // Loop over all elements
  for(int e=0; e<getNumElements(); e++)
  {
    // Data term
    if(outside[e])
    {
      //data += round(data_w*pow((m_values[e]-source_v)>0?(m_values[e]-source_v):0,2));
      data += m_DataTermIn[e];
    }else
    {
      //data += round(data_w*pow((dest_v-m_values[e])>0?(dest_v-m_values[e]):0,2));
      data += m_DataTermOut[e];
    }
    // Smoothness term
    // loop over neighbors of e
    for(int ni = 0;ni<(d+1);ni++)
    {
      const int n = m_elemNeighbors[e].getV(ni);
      // Only consider real neighbors
      if(n != -1)
      {
        // Only handle eminating edges
        if( !outside[e] && outside[n] )
        {
          //smoothness += (round(smooth_w*exp(-fabs(m_values[e]-m_values[n])))+1);
          assert(m_SmoothnessTerm.count(std::pair<int,int>(e,n)) == 1);
          smoothness += m_SmoothnessTerm[std::pair<int,int>(e,n)];
        }
      }
    }
  }

  return data+smoothness;
}

template <unsigned int d>
double Peeling<d>::cut_capacity_change(
  const BoolVector & outside, 
  const int e, 
  const bool out)
{
  assert(e>=0);
  assert(e<outside.size());
  // zero-change
  if(outside[e] == out)
  {
    return 0;
  }
  double change = 0;

  // Compute data contribution
  if(out)
  {
    // if out then we're cutting the edge from source to e and adding back the
    // edge from e to dest
    //change += round(data_w*pow((m_values[e]-source_v)>0?(m_values[e]-source_v):0,2));
    //change -= round(data_w*pow((dest_v-m_values[e])>0?(dest_v-m_values[e]):0,2));
    change += m_DataTermIn[e];
    change -= m_DataTermOut[e];
  }else
  {
    // if in then we're cutting the edge from e to dest and adding back the
    // edge from source to e
    //change += round(data_w*pow((dest_v-m_values[e])>0?(dest_v-m_values[e]):0,2));
    //change -= round(data_w*pow((m_values[e]-source_v)>0?(m_values[e]-source_v):0,2));
    change += m_DataTermOut[e];
    change -= m_DataTermIn[e];
  }
  // loop over neighbors of e
  for(int ni = 0;ni<(d+1);ni++)
  {
    const int n = m_elemNeighbors[e].getV(ni);
    // Only consider real neighbors
    if(n != -1)
    {
      if( out != outside[n] )
      {
        // new assignment creates boundary between e and n: one of these edges
        // will be in -> out 
        //change += (round(smooth_w*exp(-fabs(m_values[e]-m_values[n])))+1);
        if(outside[n]){
          // New boundary edge from e to n
          change += m_SmoothnessTerm[std::pair<int,int>(e,n)];
        }else
        {
          // New boundary edge from n to e
          change += m_SmoothnessTerm[std::pair<int,int>(n,e)];
        }
      }else
      {
        // new assignment removes boundary between e and n: one of these edges
        // was in -> out
        //change -= (round(smooth_w*exp(-fabs(m_values[e]-m_values[n])))+1);
        if(outside[n]){
          // Used to be boundary edge from e to n
          change -= m_SmoothnessTerm[std::pair<int,int>(e,n)];
        }else
        {
          // Used to be boundary edge from n to e
          change -= m_SmoothnessTerm[std::pair<int,int>(n,e)];
        }
      }
    }
  }
  return change;
}

// Given a list of element indices, find all those among them that are not
// outside (i.e. inside). If element e is seen then consider outside according
// to out rather than outside[e]
//
// Inputs:
//   outside  current indicator of outside elements
//   Ele list of element indices
//   out_query  query value
//   e  id of special index
//   out  indicator of e's outside status
// Returns subset of Ele that are inside
static std::vector<int> find_match(
  const BoolVector & outside,
  const std::vector<int> & Ele,
  const bool out_query,
  const int e,
  const bool out)
{
  using namespace std;
  vector<int> IEle;
  // Only keep those that are inside
  for(vector<int>::const_iterator nit = Ele.begin(); nit!=Ele.end(); nit++)
  {
    if(e == *nit)
    {
      if(!out)
      {
        IEle.push_back(*nit);
      }
    }else if(!outside[*nit])
    {
      IEle.push_back(*nit);
    }
  }
  return IEle;
}

template<unsigned int d> 
bool Peeling<d>::is_legal(
  const BoolVector & outside, const int e, const bool out)
{
  using namespace std;
  // Handle legality in terms of the facet constraints
  bool facet_legal = true;
  // it's always legal to put e inside
  // Q: Certainly it's always legal in terms of the facet constraints. But is
  // it always legal in terms of the manifoldness constraint?
  // A: No!
  if(out)
  {
    // check nieghbors
    for(int ni = 0;ni<(d+1);ni++)
    {
      const int n = m_elemNeighbors[e].getV(ni);
        // Only a concern if n is outside or if boundary facet
        if(n==-1 || outside[n])
        {
          // Is facet between e and n a constrained facet?
          if(m_elemConstrNeighbors[e].getV(ni))
          {
            // not legal
            facet_legal = false;
            break;
          }
        }
    }
  }

  // TODO: leave early if not facet_legal

  bool manifold_legal = true;
  //// Need to check each d-2 simplex (3D: edge, 2D: vertex)

  //// loop over d-2 simplices

  // Q: Perhaps it should be "legal" if edge/vertex is already
  // non-manifold? Especially if we're adding more elements.
  // A: Maybe. But might have to be careful if calling improve() during
  // project_single_cc()
  if(d == 3)
  {
    for(int i = 0;i<6&&manifold_legal;i++)
    {
      const int u = m_elements[e].getV(EDGES_3[i][0]);
      const int v = m_elements[e].getV(EDGES_3[i][1]);
      // Collect all elements incident on {u,v}
      const vector<int> Nuv = incident(u,v);
      for(int b = 0;b<2;b++)
      {
        vector<int> INuv = find_match(outside,Nuv,b,e,out);
        manifold_legal &= single_cc(INuv);
      }
    }
  }

  // Need to check each d-3 simplex (3D: vertex, 2D: NA)
  for(int i = 0;i<(d+1)&&manifold_legal;i++)
  {
    const vector<int> Ni = incident(m_elements[e].getV(i));
    //cout<<"Ni: ";print(Ni);
    for(int b = 0;b<2;b++)
    {
      vector<int> INi = find_match(outside,Ni,b,e,out);
      //cout<<"INi: ";print(INi);
      bool single = single_cc(INi);
      manifold_legal &= single;
    }
    //if(facet_legal && !single)
    //{
    //  //cout<<"  % "<<e+1<<" "<<out<<endl;
    //  //cout<<"  ele("<<e+1<<")["<<i<<"]: "<<m_elements[e].getV(i)+1<<endl;
    //  //mexErrMsgTxt("  % !single");
    //}
  }

  return facet_legal && manifold_legal;
}

template <unsigned int d>
std::vector<int> Peeling<d>::incident(const int u, const int v)
{
  assert(d==3);
  Tuple<2> edge;
  edge.setV(0,u);
  edge.setV(1,v);
  edge.sort();
  assert(m_edgeElementMap.count(edge)==1);
  return m_edgeElementMap[edge];
}

template <unsigned int d>
std::vector<int> Peeling<d>::incident(const int i)
{
  assert(m_vertexElementMap.count(i)==1);
  return m_vertexElementMap[i];
}

template <unsigned int d> 
bool Peeling<d>::single_cc(const std::vector<int> & Ele)
{
  using namespace std;
  // Annoying base case
  // Strictly this is not a single connected component but we *MUST* return
  // true
  if(Ele.size() == 0)
  {
    return true;
  }
  // If we have 1 CC then a BFS will reach everybody

  // map of elements in Ele
  map<int,bool> in_Ele;
  map<int,bool> seen;
  for(vector<int>::const_iterator eit = Ele.begin(); eit!=Ele.end(); eit++)
  {
    in_Ele[*eit] = true;
    seen[*eit] = false;
  }

  int seen_count = 0;
  queue<int> Q;
  // Arbitrary root
  Q.push(Ele[0]);
  while(!Q.empty())
  {
    const int e = Q.front();
    Q.pop();
    if(seen[e])
    {
      // already seen
      continue;
    }
    seen[e] = true;
    seen_count++;
    // loop over neighbors of e
    for(int ni = 0;ni<(d+1);ni++)
    {
      const int n = m_elemNeighbors[e].getV(ni);
      // Only consider real neighbors in Ele
      if(n != -1 && in_Ele.count(n)==1)
      {
        assert(in_Ele[n]);
        // Only keep unseen neighbors
        if(!seen[n])
        {
          Q.push(n);
        }
      }
    }
  }
  return seen_count == Ele.size();
}


template<unsigned int d> bool Peeling<d>::testConstrFacetsPresence()
{
	const int numConstrFacets = (int)m_constrFacets.size();
	bool ret = true;
	for (int i=0; i<numConstrFacets; i++)
	{
		Tuple<d> facet = m_constrFacets[i];
		facet.sort();
		typename FacetElementMap::iterator it = m_facetElementMap.find(facet);
		if (it == m_facetElementMap.end()) 
		{
			if (d == 2) TRACE("Constrained edge no. %i (%i,%i) is not a triangle edge\n", i, m_constrFacets[i].getV(0), m_constrFacets[i].getV(1));
			else if (d == 3) TRACE("Constrained triangle no. %i (%i,%i,%i) is not a tetrahedron triangle\n", i, m_constrFacets[i].getV(0), m_constrFacets[i].getV(1), m_constrFacets[i].getV(2));
			else TRACE("error, unsupported dimension (%i)\n");
			ret = false;
		}
	}
	return ret;
}

template<unsigned int d> bool Peeling<d>::prepareFacetElementMap()
{
	for (int e=0; e<getNumElements(); e++)
	{
		for (int f=0; f<d+1; f++)
		{
			Tuple<d> facet, facetUnsorted;
			for (int i=0; i<d; i++) facet.setV(i, m_elements[e].getV((f+i)%(d+1)));			
			facetUnsorted = facet;
			facet.sort();
			typename FacetElementMap::iterator it = m_facetElementMap.find(facet);
			if (it == m_facetElementMap.end())
			{	// not found
				m_facetElementMap[facet] = Tuple<2>(e, -1);
			}
			else
			{
				Tuple<2> record = it->second;
				if (record.getV(0) < 0) 
				{
					TRACE("prepareFacetElementMap -- empty entry in m_facetElementMap\n");
					return false;
				}
				if (record.getV(1) >= 0) 
				{
					if (d == 2)	TRACE("prepareFacetElementMap -- edge (%i, %i) belongs to more than 2 triangles\n", facetUnsorted.getV(0), facetUnsorted.getV(1));
					else if (d == 3) TRACE("prepareFacetElementMap -- triangle (%i, %i, %i) belongs to more than 2 tets: %i, %i, %i\n", facetUnsorted.getV(0), facetUnsorted.getV(1), facetUnsorted.getV(2), 
						record.getV(0), record.getV(1), e);
					else TRACE("prepareFacetElementMap -- not supported dimension (%i)\n", d);
					
					return false;
				}
				record.setV(1, e);
				m_facetElementMap[facet] = record;
			}
		}
	}

	return true;
}

template<unsigned int d> 
void Peeling<d>::prepareEdgeElementMap()
{
  assert(d==3);
  // loop over d-2 simplices
  for (int e=0; e<getNumElements(); e++)
  {
    for(int i = 0;i<6;i++)
    {
      Tuple<2> edge;
      edge.setV(0,m_elements[e].getV(EDGES_3[i][0]));
      edge.setV(1,m_elements[e].getV(EDGES_3[i][1]));
      edge.sort();
      m_edgeElementMap[edge].push_back(e);
    }
  }
}

template<unsigned int d> 
void Peeling<d>::prepareVertexElementMap()
{
  for (int e=0; e<getNumElements(); e++)
  {
    for(int i = 0;i<d+1;i++)
    {
      m_vertexElementMap[m_elements[e].getV(i)].push_back(e);
    }
  }
}

template<unsigned int d> void Peeling<d>::prepareElemNeighbors()
{
	// relies on m_facetElementMap
	Tuple<d+1> invalid;
	for (int i=0; i<d+1; i++) invalid.setV(i, -2);

	m_elemNeighbors.assign(getNumElements(), invalid);
	m_elemConstrNeighbors.assign(getNumElements(), invalid);

	for (int e=0; e<getNumElements(); e++)
	{
		for (int f=0; f<d+1; f++)
		{
			Tuple<d> facet;
			for (int i=0; i<d; i++) facet.setV(i, m_elements[e].getV((f+i)%(d+1)));
			facet.sort();
			typename FacetElementMap::iterator it = m_facetElementMap.find(facet);
			if (it == m_facetElementMap.end())
			{	// not found
				assert(false);
				mexErrMsgTxt("prepareElemNeighbors -- internal bug 1");				
			}
			else
			{
				Tuple<2> record = it->second;
				if (record.getV(0) == e)
				{
					m_elemNeighbors[e].setV(f, record.getV(1));
				}
				else if (record.getV(1) == e)
				{
					m_elemNeighbors[e].setV(f, record.getV(0));
				}
				else
				{
					assert(false);
					mexErrMsgTxt("prepareElemNeighbors -- internal bug 2");				
				}
			}

			typename FacetSet::iterator it2 = m_constrFacetSet.find(facet);
			if (it2 == m_constrFacetSet.end()) 
			{
				m_elemConstrNeighbors[e].setV(f, 0); // not constrained
			}
			else 
			{
				m_elemConstrNeighbors[e].setV(f, 1);
			}
		}
	}
}

template<unsigned int d> void Peeling<d>::prepareConstrFacetSet()
{
	m_constrFacetSet.clear();
	const int numConstrFacets = (int)m_constrFacets.size();	

	for (int i=0; i<numConstrFacets; i++)
	{
		Tuple<d> facet = m_constrFacets[i];
		facet.sort();
		m_constrFacetSet.insert(facet);
	}
}

template<unsigned int d> DoubleIntPairVector Peeling<d>::sortElementsByValues()
{
	DoubleIntPairVector sortedElements;
	sortedElements.clear();
	for (int i=0; i<getNumElements(); i++)
	{		
		sortedElements.push_back(DoubleIntPair(m_values[i], i));
	}
	std::sort(sortedElements.begin(), sortedElements.end());
	return sortedElements;
}

template<unsigned int d> int Peeling<d>::connectedComponentInside(const BoolVector &outside, int start)
{
	BoolVector visited; visited.assign(getNumElements(), false);
	typedef std::queue<int> IntQueue;
	IntQueue q;

	assert(!outside[start]); // otherwise doesn't make sense
	q.push(start);
	
	int count = 0;
	while (!q.empty())
	{		
		int node = q.front();
		q.pop();
		assert(node >= 0 && node < getNumElements());
		if (visited[node]) continue; // this means we've discovered the vertex again through another path
		assert(!outside[node]);
		count++;
		visited[node] = true;
		
		for (int i=0; i<d+1; i++)
		{ // loop over neighbors:
			const int neighbor = m_elemNeighbors[node].getV(i);
			if (neighbor < 0) continue;
			assert(neighbor >= 0 && neighbor < getNumElements());			
			if (visited[neighbor] || outside[neighbor]) continue;

			q.push(neighbor);
		}
	}

	return count;
}

// explicit instantiation
template class Peeling<2>;
template class Peeling<3>;
#pragma GCC diagnostic pop
