#include "post_graph_cut.h"
#include "Peeling.h"

#include <igl/matlab/mexStream.h>
#include <iostream>
#include <mex.h>

void post_graph_cut(
  const int d,
  const int numElements,
  const double * pElements,
  const double * pValues,
  const int numFacets,
  const double * pFacets,
  const std::map<std::pair<int,int>,double> & pSmoothnessTerm,
  const std::vector<double> & pDataTermIn,
  const std::vector<double> & pDataTermOut,
  BoolVector & outside)
{
  using namespace std;
  switch(d)
  {
    case 2:
    {
      Peeling<2> peeler;
      //peeler.init(numElements, pElements, pValues, numFacets, pFacets);
      peeler.init(numElements, pElements, pValues, numFacets, pFacets,pSmoothnessTerm,pDataTermIn,pDataTermOut);	  
      cout<<"cut-capacity before: "<<peeler.cut_capacity(outside)<<endl;
      peeler.project(outside);
      cout<<"cut-capacity after project: "<<peeler.cut_capacity(outside)<<endl;
      queue<int> dirtyQ;
      for(int e = 0;e<numElements;e++)
      {
        dirtyQ.push(e);
      }
      const int flips = peeler.improve(outside,dirtyQ,true);
      cout<<"final flips (should be 0): "<<flips<<endl;
      cout<<"cut-capacity after improve: "<<peeler.cut_capacity(outside)<<endl;
      break;
    }
    case 3:
    {
      Peeling<3> peeler;
      //peeler.init(numElements, pElements, pValues, numFacets, pFacets);	  
      peeler.init(numElements, pElements, pValues, numFacets, pFacets,pSmoothnessTerm,pDataTermIn,pDataTermOut);	  
      cout<<"cut-capacity before: "<<peeler.cut_capacity(outside)<<endl;
      peeler.project(outside);
      cout<<"cut-capacity after project: "<<peeler.cut_capacity(outside)<<endl;
      queue<int> dirtyQ;
      for(int e = 0;e<numElements;e++)
      {
        dirtyQ.push(e);
      }
      const int flips = peeler.improve(outside,dirtyQ,true);
      cout<<"final flips (should be 0): "<<flips<<endl;
      cout<<"cut-capacity after improve: "<<peeler.cut_capacity(outside)<<endl;
      break;
    }
  }
}

void mexFunction(
  int nlhs, mxArray *plhs[], 
  int nrhs, const mxArray *prhs[])
{
  igl::mexStream mout;        
  std::streambuf *outbuf = std::cout.rdbuf(&mout);
  mexPrintf("Compiled at %s on %s\n",__TIME__,__DATE__);

  // Parse RHS
  if (nrhs != 6) mexErrMsgTxt("Expecting 6 input arguments: elements (numElements x (d+1), values (numElements x 1), constrained facets (numFacets x d), initial outside (numElements x 1), smoothness weights matrix (sparse numElements x numElements), data weights matrix (sparse numElements x 2)");

  int numElements = (int)mxGetM(prhs[0]);
  int d = (int)mxGetN(prhs[0]);
  if (d != 3 && d != 4) mexErrMsgTxt("Dimension d+1 has to be 3 or 4");  

  int tmp;
  tmp = (int)mxGetM(prhs[1]);
  if (tmp != numElements) mexErrMsgTxt("Numbers of rows of first two arguments have to match");
  if (mxGetN(prhs[1]) != 1) mexErrMsgTxt("Second argument has to be column vector");  

  int numFacets = (int)mxGetM(prhs[2]);
  if (mxGetN(prhs[2]) != d-1) mexErrMsgTxt("Third argument has to be numEdges x (d-1) matrix");  

  tmp = (int)mxGetM(prhs[3]);
  if (tmp != numElements) mexErrMsgTxt("Numbers of rows of first and fourth arguments have to match");
  if (mxGetN(prhs[3]) != 1) mexErrMsgTxt("Fourth argument has to be column vector");  

  if (!mxIsDouble(prhs[0]) || !mxIsDouble(prhs[1]) || !mxIsDouble(prhs[2]) || !mxIsDouble(prhs[3])) mexErrMsgTxt("All matrices have to be doubles");

  if(!mxIsSparse(prhs[4]) || !mxIsSparse(prhs[5])) mexErrMsgTxt("Fifth and sixth arguments should be sparse");
  tmp = (int)mxGetM(prhs[4]);
  if (tmp != numElements) mexErrMsgTxt("Numbers of rows of first and fifth arguments have to match");
  tmp = (int)mxGetN(prhs[4]);
  if (tmp != numElements) mexErrMsgTxt("Numbers of rows of first and number of columns fifth arguments have to match");
  tmp = (int)mxGetM(prhs[5]);
  if (tmp != numElements) mexErrMsgTxt("Numbers of rows of first and sixth arguments have to match");
  if (mxGetN(prhs[5]) != 2) mexErrMsgTxt("Sixth argument has to be a 2-column matrix");  


  double *pElements = mxGetPr(prhs[0]);
  double *pValues = mxGetPr(prhs[1]);
  double *pFacets = mxGetPr(prhs[2]);
  double *pInitialOutside = mxGetPr(prhs[3]);
  const mxArray *pSmoothnessTerm = prhs[4];
  const mxArray  *pDataTerm = prhs[5];

  // Build smoothness term as stl map
  std::map<std::pair<int,int>,double> SmoothnessTerm;
  {
    const double  *pr = mxGetPr(pSmoothnessTerm);
    const mwIndex *ir = mxGetIr(pSmoothnessTerm);
    const mwIndex *jc = mxGetJc(pSmoothnessTerm);
    // traverse the adjacency matrix and add n-links
    unsigned int i, j, k;
    double v;
    for (j = 0; j < numElements; j++)
    {
      // this is a simple check whether there are non zero
      // entries in the j'th column
      if (jc[j] == jc[j+1])
      {
        continue; // nothing in this column
      }
      for (k = jc[j]; k <= jc[j+1]-1; k++)
      {
        i = ir[k];
        v = pr[k];
        //mexPrintf("non zero entry: (%d,%d) = %.2f\n", i+1, j+1, v);
        if(!(SmoothnessTerm.count(std::pair<int,int>(i,j))==0)) mexErrMsgTxt("SmoothnessTerm.count(std::pair<int,int>(i,j))==0");
        SmoothnessTerm[std::pair<int,int>(i,j)] = v;
      }
    }
  }

  // Build data term as 2 stl vectors
  // Is zero really the best initializer here? This means empty elements are
  // zero-weighted edges.
  std::vector<double> DataTermIn(numElements,0);
  std::vector<double> DataTermOut(numElements,0);
  {
    const double  *pr = mxGetPr(pDataTerm);
    const mwIndex *ir = mxGetIr(pDataTerm);
    const mwIndex *jc = mxGetJc(pDataTerm);
    // traverse the adjacency matrix and add n-links
    unsigned int i, j, k;
    double v;
    for (j = 0; j < 2; j++)
    {
      // this is a simple check whether there are non zero
      // entries in the j'th column
      if (jc[j] == jc[j+1])
      {
        continue; // nothing in this column
      }
      for (k = jc[j]; k <= jc[j+1]-1; k++)
      {
        i = ir[k];
        v = pr[k];
        //mexPrintf("non zero entry: (%d,%d) = %.2f\n", i+1, j+1, v);
        if(j==0)
        {
          DataTermOut[i] = v;
        }else
        {
          DataTermIn[i] = v;
        }
      }
    }
  }


  // Initialize outside
  BoolVector outside(numElements);
  for (int e=0; e<numElements; e++) outside[e] = pInitialOutside[e];	  
  post_graph_cut(d-1,numElements,pElements,pValues,numFacets,pFacets,SmoothnessTerm,DataTermIn,DataTermOut,outside);

  if(nlhs > 0)
  {
    plhs[0] = mxCreateDoubleMatrix(numElements,1, mxREAL);
    double *pRes = mxGetPr(plhs[0]);
    for (int e=0; e<numElements; e++) pRes[e] = outside[e];	  
  }

  std::cout.rdbuf(outbuf);
}
